//==========================================================================================================================================================================================================
// HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER      HEADER
//==========================================================================================================================================================================================================
#include <list>
#include <iostream>
#include <vector>

#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include "Math.hpp"
#include <windows.h>

void CoreSetup();
void CoreUpdate();
void BuildWorld();
void ShowFPS();
void DrawLevel(Vector3D);
void UpdateScreen();
//==========================================================================================================================================================================================================
// END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER      END_HEADER
//==========================================================================================================================================================================================================

//==========================================================================================================================================================================================================
// DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA      DATA
//==========================================================================================================================================================================================================
struct Rotator
{
    Single      Pitch;
    Single      Yaw;
    Single      Roll;
};
struct Mesh
{
    String      Name;
    Int32       ID;
    Int64       Size;
    Vector3D*   Voxel;
    Vector3D    Location;
    Rotator     Rotation;
};
struct Particle
{
    __m128 Coord;
    Single Color;
};

HWND    WindowHandle;
HDC     WindowDC;

Rotator     CameraRotation, CameraArmRotation;
Matrix3D    CameraMat, CameraArmMat;

Int32   CountFPS;
UInt64  DeltaFPS;
size_t  SCREEN_WIDTH  = 1280;
size_t  SCREEN_HEIGHT = 720;
std::vector<UInt32> Screen(SCREEN_WIDTH * SCREEN_HEIGHT);
std::vector<UInt32> Monoscope(SCREEN_WIDTH * SCREEN_HEIGHT);

Particle* deviceInputData;
UInt32*   deviceOutputData;

Int32   SpaceDim = 1000;
std::vector<Particle> Space3D(SpaceDim * SpaceDim * SpaceDim);
std::list<Mesh>       MeshDB;
//==========================================================================================================================================================================================================
// END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA      END_DATA
//==========================================================================================================================================================================================================

//==========================================================================================================================================================================================================
// CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE      CODE
//==========================================================================================================================================================================================================
void    CoreEntry()
{
    CoreSetup();
    CoreUpdate();
}

void    CoreSetup()
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleTextAttribute(hConsole, FOREGROUND_GREEN | FOREGROUND_INTENSITY);

    #pragma omp parallel for
    for (Int32 X = 0; X < SpaceDim; X++)
    {
        for (Int32 Y = 0; Y < SpaceDim; Y++)
        {
            for (Int32 Z = 0; Z < SpaceDim; Z++)
            {
                Int32 Ratio = (SpaceDim / 2);
                Int32 Index = X * SpaceDim * SpaceDim + Y * SpaceDim + Z;   // X + SpaceDim * (Y + SpaceDim * Z);

                Space3D[Index].Coord.m128_f32[0] = (Single)X - Ratio;
                Space3D[Index].Coord.m128_f32[1] = (Single)Y - Ratio;
                Space3D[Index].Coord.m128_f32[2] = (Single)Z - Ratio;
            }
        }
    }
}

void    CoreUpdate()
{
    // Allocate data on GPU memory
    cudaMalloc(&deviceInputData , Space3D.size()   * sizeof(Particle));

    // Copy to GPU memory
    cudaMemcpy(deviceInputData, Space3D.data(), Space3D.size() * sizeof(Particle), cudaMemcpyHostToDevice);

    while (true)
    {
        BuildWorld();
        ShowFPS();
    }
}
//==========================================================================================================================================================================================================
// END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE      END_CODE
//==========================================================================================================================================================================================================

//==========================================================================================================================================================================================================
// FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION      FUNCTION
//==========================================================================================================================================================================================================
__global__ void GPUKernel(UInt32* Screen, Particle* Space3D, Matrix3D CameraMat, Matrix3D CameraArmMat, int SCREEN_WIDTH, int SCREEN_HEIGHT)
{
    Vector3D CameraLocation, CameraArmLocation;
    Vector3D Dot, Res;

    Int32 tidX = blockIdx.x * blockDim.x + threadIdx.x;
    Dot.Col = Space3D[tidX].Coord;

    // 1) Move the mesh and modify origin of rotation
    // 2) Move the mesh and keep   origin of rotation
    //
    // 1)
    //    Rotate the mesh around the origin of rotation
    // 2)

    for (Int32 Index = 0; Index < 4; Index++)
    {
        // Vector.Add
        Res.Arr[Index] += CameraLocation.Arr[Index];

        // Vector.Transform
        Res.Arr[Index] += (Dot.Arr[0] * CameraMat.Arr[Index][0]);
        Res.Arr[Index] += (Dot.Arr[1] * CameraMat.Arr[Index][1]);
        Res.Arr[Index] += (Dot.Arr[2] * CameraMat.Arr[Index][2]);
        Res.Arr[Index] += (Dot.Arr[3] * CameraMat.Arr[Index][3]);

        // Vector.Add
        Res.Arr[Index] += CameraArmLocation.Arr[Index];

        // Vector.Transform
        Res.Arr[Index] += (Dot.Arr[0] * CameraArmMat.Arr[Index][0]);
        Res.Arr[Index] += (Dot.Arr[1] * CameraArmMat.Arr[Index][1]);
        Res.Arr[Index] += (Dot.Arr[2] * CameraArmMat.Arr[Index][2]);
        Res.Arr[Index] += (Dot.Arr[3] * CameraArmMat.Arr[Index][3]);
    }
    Dot.Col = Res.Col;

    Int32   X, Y;

    if (Dot.Z != 0)
    {
        X = (Int32)(Dot.X * 500 / Dot.Z);
        Y = (Int32)(Dot.Y * 500 / Dot.Z);
    }
    else
    {
        X = (Int32)(Dot.X * 500);
        Y = (Int32)(Dot.Y * 500);
    }

    Int32 LimX    = (SCREEN_WIDTH  / 2);
    Int32 LimY    = (SCREEN_HEIGHT / 2);
    Int32 Repere  = (SCREEN_WIDTH  / 2) + (SCREEN_HEIGHT * SCREEN_WIDTH / 2);
    Int32 OffsetPixel = Repere - (Y * SCREEN_WIDTH + X);

    if (X < LimX && X > -LimX && Dot.Z >= 0 &&
        Y < LimY && Y > -LimY && OffsetPixel < SCREEN_WIDTH * SCREEN_HEIGHT)
    {
        Screen[OffsetPixel] = 0x0000FF00;   // 0x00RRGGBB
    }
}
void    BuildWorld()
{
    Matrix3D    Matrix;
    Matrix3D    Pitch, Yaw, Roll;

    {/*************************************************************************/}
     /*************************************************************************/
     /**/  Pitch        = Matrix.CreateRotationX(CameraRotation.Pitch++);   /**/
     /**/  Yaw          = Matrix.CreateRotationY(CameraRotation.Yaw);       /**/
     /**/  Roll         = Matrix.CreateRotationZ(CameraRotation.Roll);      /**/
     /**/  CameraMat    = Roll * Yaw * Pitch;                               /**/
     /**/                                                                   /**/
     /**/  Pitch        = Matrix.CreateRotationX(CameraArmRotation.Pitch);  /**/
     /**/  Yaw          = Matrix.CreateRotationY(CameraArmRotation.Yaw);    /**/
     /**/  Roll         = Matrix.CreateRotationZ(CameraArmRotation.Roll);   /**/
     /**/  CameraArmMat = Roll * Yaw * Pitch;                               /**/
     /*************************************************************************/
    {/*************************************************************************/}

    // Allocate data on GPU memory
    cudaMalloc(&deviceOutputData, Monoscope.size() * sizeof(UInt32));

    // Run a kernel
    GPUKernel<<<1000000, 1000>>>(deviceOutputData, deviceInputData, CameraMat, CameraArmMat, SCREEN_WIDTH, SCREEN_HEIGHT);

    // Copy data to CPU memory
    cudaMemcpy(Screen.data(), deviceOutputData, Monoscope.size() * sizeof(UInt32), cudaMemcpyDeviceToHost);
    
    // Free data on GPU memory
    cudaFree(deviceOutputData);

    InvalidateRect(WindowHandle, NULL, FALSE);
    //RedrawWindow(WindowHandle, NULL, NULL, RDW_INVALIDATE);
}
void    ShowFPS()
{
    CountFPS++;

    if (GetTickCount64() > DeltaFPS + 1000)
    {
        printf("FPS = %d\n", CountFPS);

        CountFPS = 0;
        DeltaFPS = GetTickCount64();
    }
}
//==========================================================================================================================================================================================================
// END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION      END_FUNCTION
//==========================================================================================================================================================================================================