#include <immintrin.h>
#include <corecrt_math_defines.h>
#include <corecrt_math.h>

typedef          __int8  Int8;
typedef          __int16 Int16;
typedef          __int32 Int32;
typedef          __int64 Int64;
typedef unsigned __int8  UInt8;
typedef unsigned __int16 UInt16;
typedef unsigned __int32 UInt32;
typedef unsigned __int64 UInt64;
typedef          char*   String;
typedef          float   Single;
typedef          double  Double;
typedef          bool    Boolean;

struct Matrix3D
{
    union
    {
        struct
        {
            Single   M11, M12, M13, M14;
            Single   M21, M22, M23, M24;
            Single   M31, M32, M33, M34;
            Single   M44, M43, M42, M41;
        };
        __m128 Row[4];
        float Arr[4][4];
    };

    friend Matrix3D     operator*(Matrix3D left, Matrix3D right)
    {
        Matrix3D Res = { 0 };
        __m128   Dst;

        for (Int32 i = 0; i < 4; i++)
        {
            for (Int32 j = 0; j < 4; j++)
            {
                Dst = _mm_mul_ps(left.Row[i], right.Row[j]);
                Dst = _mm_hadd_ps(Dst, Dst);
                Dst = _mm_hadd_ps(Dst, Dst);

                Res.Arr[i][j] += Dst.m128_f32[0];

                //result.Arr[i][j] += left.Arr[i][0] * right.Arr[0][j];
                //result.Arr[i][j] += left.Arr[i][1] * right.Arr[1][j];
                //result.Arr[i][j] += left.Arr[i][2] * right.Arr[2][j];
                //result.Arr[i][j] += left.Arr[i][3] * right.Arr[3][j];
            }
        }

        return (Res);
    }

    Single      DEG(Single Radian)
    {
        return (Single)(Radian * (180.0F / M_PI));
    }
    Single      RAD(Single Degree)
    {
        return (Single)(M_PI * Degree / 180.0F);
    }
    Matrix3D    CreateRotationX(Single Angle)
    {
        Angle = RAD(Angle);

        Matrix3D Matrix = { 1 ,     0      ,      0        , 0 ,
                            0 , cosf(Angle) , -sinf(Angle) , 0 ,
                            0 , sinf(Angle) ,  cosf(Angle) , 0 ,
                            0 ,     0      ,      0        , 1 };


        return (Matrix);
    }
    Matrix3D    CreateRotationY(Single Angle)
    {
        Angle = RAD(Angle);

        Matrix3D Matrix = { cosf(Angle) , 0 , -sinf(Angle) , 0 ,
                                0       , 1 ,      0       , 0 ,
                            sinf(Angle) , 0 ,  cosf(Angle) , 0 ,
                                0       , 0 ,      0       , 1 };


        return (Matrix);
    }
    Matrix3D    CreateRotationZ(Single Angle)
    {
        Angle = RAD(Angle);

        Matrix3D Matrix = { cosf(Angle) , -sinf(Angle) , 0 , 0 ,
                            sinf(Angle) ,  cosf(Angle) , 0 , 0 ,
                                 0      ,       0      , 1 , 0 ,
                                 0      ,       0      , 0 , 1 };

        return (Matrix);
    }
};
struct Vector3D
{
    union
    {
        struct
        {
            Single  X;
            Single  Y;
            Single  Z;
            Single  W;
        };

        __m128  Col;
        Single  Arr[4];
    };

    Vector3D    Add(Vector3D left, Vector3D right)
    {
        Vector3D res;

        res.Col = _mm_add_ps(left.Col, right.Col);

        return (res);
    }
    Vector3D    Sub(Vector3D left, Vector3D right)
    {
        Vector3D res;

        res.Col = _mm_sub_ps(left.Col, right.Col);

        return (res);
    }
    Vector3D    Sub(Vector3D left, Single value)
    {
        Vector3D res;

        res.Col = _mm_sub_ps(left.Col, _mm_broadcast_ss(&value));

        return (res);
    }
    Vector3D    Mul(Vector3D left, Vector3D right)
    {
        Vector3D res;

        res.Col = _mm_mul_ps(left.Col, right.Col);

        return (res);
    }
    Vector3D    Div(Vector3D left, Vector3D right)
    {
        Vector3D res;

        res.Col = _mm_div_ps(left.Col, right.Col);

        return (res);
    }
    Vector3D    Transform(Vector3D vector, Matrix3D matrix)
    {
        Vector3D Res = { 0 };
        __m128   Dst;

        for (int i = 0; i < 4; i++)
        {
            Dst = _mm_mul_ps(matrix.Row[i], vector.Col);
            Dst = _mm_hadd_ps(Dst, Dst);
            Dst = _mm_hadd_ps(Dst, Dst);

            Res.Arr[i] += Dst.m128_f32[0];
        }

        return (Res);
    }
};
struct Math3D
{
    Boolean     Almost(Single value1, Single value2)
    {
        return (abs(value1 - value2) <= 10);
    }
    Single      DEG(Single Radian)
    {
        return (Single)(Radian * (180.0F / M_PI));
    }
    Single      RAD(Single Degree)
    {
        return ((Single)M_PI * Degree / 180.0F);
    }
    Single      NEWT(Single Gram)
    {
        return (Gram * 0.00980665002863885F);
    }
    Single      GRAM(Single Newton)
    {
        return (Newton * 101.971621F);
    }
    Single      LimitAngle(Single Angle)
    {
        if (Angle >= 180)
        {
            Angle -= 360;
        }
        else if (Angle <= -180)
        {
            Angle += 360;
        }

        return (Angle);
    }
    Single      Clamp(Single X, Single Min, Single Max)
    {
        if (X < Min)
        {
            return (Min);
        }
        else
        {
            if (X < Max)
            {
                return (X);
            }
            else
                return (Max);
        }
    }
    //Vector3D    InterpTo(Vector3D Current, Vector3D Target, Single InterpSpeed)
    //{
    //    // 0.90 < InterpSpeed < 1

    //    Vector3D Delta;
    //    {
    //        Delta.X = (Target.X - Current.X);
    //        Delta.Y = (Target.Y - Current.Y);
    //    }

    //    Vector3D DeltaMove;
    //    {
    //        DeltaMove.X = (Target.X - Current.X) * InterpSpeed;
    //        DeltaMove.Y = (Target.Y - Current.Y) * InterpSpeed;
    //    }

    //    Target.X = (Current.X + DeltaMove.X);
    //    Target.Y = (Current.Y + DeltaMove.Y);

    //    return (Target);
    //}
};